import abc
import pandas as pd
import pygit2 as git


class History(abc.ABC):

    def __init__(self, repository: git.Repository, branch: str = "master"):
        self.repo = repository
        self.branch = branch
        self.cache = None
        self.mailmap = git.Mailmap.from_repository(self.repo)

    def map_signature(self, signature: git.Signature):
        # the unmapped email is used on purpose
        email = signature.email
        try:
            mapped_signature = self.mailmap.resolve_signature(signature)
            name = mapped_signature.name
        except ValueError:
            name = signature.name
            if not name:
                name = "Empty Empty"
                # warnings.warn(f"{str(e)}. Name will be replaced with '{name}'")
            if not email:
                email = "empty@empty.empty"
                # warnings.warn(f"{str(e)}. Email will be replaced with '{email}'")
        return name, email

    def as_dataframe(self):
        data = self.fetch()
        return pd.DataFrame(data)

    @abc.abstractmethod
    def fetch(self):
        pass


class WholeHistory(History):
    def fetch(self):
        repo_walker = self.repo.walk(self.repo.head.target, git.GIT_SORT_TOPOLOGICAL)
        records = []
        for commit in repo_walker:
            author_name, author_email = self.map_signature(commit.author)

            insertions, deletions = 0, 0
            if len(commit.parents) == 0:  # initial commit
                st = commit.tree.diff_to_tree(swap=True).stats
                insertions, deletions = st.insertions, st.deletions
            elif len(commit.parents) == 1:
                parent_commit = commit.parents[0]
                st = self.repo.diff(parent_commit, commit).stats
                insertions, deletions = st.insertions, st.deletions
            # case len(commit.parents) > 1 corresponds to a merge commit
            # merge commits are ignored: changes in merge commits are normally because of integration issues

            records.append({'commit_sha': commit.hex[:7],
                            'author_name': author_name,
                            'author_email': author_email,
                            'author_tz_offset': commit.author.offset,
                            'author_timestamp': commit.author.time,
                            'insertions': insertions,
                            'deletions': deletions})
        return records


class LinearHistory(History):
    def fetch(self):
        repo_walker = self.repo.walk(self.repo.head.target, git.GIT_SORT_TOPOLOGICAL)
        repo_walker.simplify_first_parent()
        records = []
        for commit in repo_walker:

            insertions, deletions = 0, 0
            if len(commit.parents) == 0:  # initial commit
                st = commit.tree.diff_to_tree(swap=True).stats
                insertions, deletions = st.insertions, st.deletions
            elif len(commit.parents) >= 1:
                parent_commit = commit.parents[0]
                st = self.repo.diff(parent_commit, commit).stats
                insertions, deletions = st.insertions, st.deletions

            records.append({'commit_sha': commit.hex[:7],
                            'committer_timestamp': commit.committer.time,
                            'files_count': len(commit.tree.diff_to_tree()),
                            'insertions': insertions,
                            'deletions': deletions})
        return records



import os
import datetime
import calendar
import collections
import json
from jinja2 import Environment, FileSystemLoader

from analysis.gitstatistics import GitStatistics
from analysis.gitrepository import GitRepository
from tools.configuration import Configuration
from tools import sort_keys_by_value_of_key
from tools import colormaps

HERE = os.path.dirname(os.path.abspath(__file__))


class HTMLReportCreator(object):
    recent_activity_period_weeks = 32
    assets_subdir = "assets"
    templates_subdir = "templates"

    def __init__(self, config: Configuration, repo_stat: GitStatistics, repository: GitRepository):
        self.path = None
        self.configuration = config
        self.assets_path = os.path.join(HERE, self.assets_subdir)
        self.git_repo_statistics = repo_stat
        self.git_repository_statistics = repository
        self.has_tags_page = config.do_process_tags()
        self._time_sampling_interval = "W"

        self.common_rendering_data = {
            "assets_path": self.assets_path,
            "has_tags_page": self.has_tags_page
        }

        templates_dir = os.path.join(HERE, self.templates_subdir)
        self.j2_env = Environment(loader=FileSystemLoader(templates_dir), trim_blocks=True)
        self.j2_env.filters['to_month_name_abr'] = lambda im: calendar.month_abbr[im]
        self.j2_env.filters['to_weekday_name'] = lambda i: calendar.day_name[i]
        self.j2_env.filters['to_ratio'] = lambda val, max_val: (float(val) / max_val) if max_val != 0 else 0
        self.j2_env.filters['to_percentage'] = lambda val, max_val: (100 * float(val) / max_val) if max_val != 0 else 0
        colors = colormaps.colormaps[self.configuration['colormap']]
        self.j2_env.filters['to_heatmap'] = lambda val, max_val: "%d, %d, %d" % colors[int(float(val) / max_val * (len(colors) - 1))]

    def set_time_sampling(self, offset: str):
        """
        :param offset: any valid string composed of Pandas' offset aliases
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        """
        self._time_sampling_interval = offset
        return self

    def _get_recent_activity_data(self):
        recent_weekly_commits = self.git_repository_statistics.\
            get_recent_weekly_activity(self.recent_activity_period_weeks)

        values = [
            {
                'x': int(self.recent_activity_period_weeks - i - 1),
                'y': int(commits)
            } for i, commits in enumerate(recent_weekly_commits)
        ]
        graph_data = {
            "xAxis": {"axisLabel": "Weeks ago"},
            "yAxis": {"axisLabel": "Commits"},
            "config": {
                "noData": "No recent activity.",
                "padData": True,
                "showXAxis": True,
                "xDomain": [self.recent_activity_period_weeks - 1, 0]
            },
            "data": [
                { 'key': "Commits", 'color': "#9400D3", 'values': values}
            ]
        }

        return graph_data

    def _bundle_assets(self):
        from distutils.dir_util import copy_tree
        # copy assets to report output folder
        assets_local_abs_path = os.path.join(self.path, self.assets_subdir)
        copy_tree(src=self.assets_path, dst=assets_local_abs_path)
        # relative path to assets to embed into html pages
        self.assets_path = os.path.relpath(assets_local_abs_path, self.path)

    def _squash_authors_history(self, authors_history, max_authors_count):
        authors = self.git_repository_statistics.authors.sort(by='commits_count').names()

        most_productive_authors = authors[:max_authors_count]
        rest_authors = authors[max_authors_count:]

        most_productive_authors_history = authors_history[most_productive_authors] \
            .asfreq(freq=self._time_sampling_interval, fill_value=0) \
            .cumsum()
        rest_authors_history = authors_history[rest_authors].sum(axis=1)\
            .asfreq(freq=self._time_sampling_interval, fill_value=0)\
            .cumsum()

        most_productive_authors_history['Others'] = rest_authors_history.values
        return most_productive_authors_history

    def create(self, path):
        self.path = path

        if self.configuration.is_report_relocatable():
            self._bundle_assets()
            self.common_rendering_data.update({
                "assets_path": self.assets_path
            })

        ###
        # General
        general_html = self.render_general_page()
        with open(os.path.join(path, "general.html"), 'w', encoding='utf-8') as f:
            f.write(general_html)

        try:
            # make the landing page for a web server
            os.symlink("general.html", os.path.join(path, "index.html"))
        except FileExistsError:
            # if symlink exists, it points to "general.html" so no need to re-create it
            # other solution would be to use approach from
            # https://stackoverflow.com/questions/8299386/modifying-a-symlink-in-python/8299671
            print("index.html already exists.")
        except OSError:
            print("index.html could not be created."
                  "On newer versions of Windows, unprivileged accounts can create symlinks only"
                  "if Developer Mode is enabled or SeCreateSymbolicLinkPrivilege privilege is granted."
                  "Otherwise, run the process as an administrator.")

        ###
        # Activity
        activity_html = self.render_activity_page()
        with open(os.path.join(path, "activity.html"), 'w', encoding='utf-8') as f:
            f.write(activity_html)

        recent_activity = self._get_recent_activity_data()

        # Commits by current year's months
        current_year = datetime.date.today().year
        current_year_monthly_activity = self.git_repository_statistics.history('m')
        current_year_monthly_activity = current_year_monthly_activity\
            .loc[current_year_monthly_activity.index.year == current_year]

        import pandas as pd
        current_year_monthly_activity = pd.Series(current_year_monthly_activity.values,
                                                  index=current_year_monthly_activity.index.month).to_dict()
        values = [
            {
                'x': imonth,
                'y': current_year_monthly_activity.get((imonth + 1), 0)
            } for imonth in range(0, 12)
        ]
        by_month = {
            "yAxis": {"axisLabel": "Commits in %d" % current_year},
            "xAxis": {"rotateLabels": -90, "ticks": len(values)},
            "config": {
                "padData": True,
                "showXAxis": True
            },
            "data": [
                { "key": "Commits", "color": "#9400D3", "values": values }
            ]
        }

        # Commits by year
        yearly_activity = self.git_repository_statistics.history('Y')
        values = [{'x': int(x.year), 'y': int(y)} for x, y in zip(yearly_activity.index, yearly_activity.values)]

        by_year = {
            "xAxis": {"rotateLabels": -90, "ticks": len(values)},
            "yAxis": {"axisLabel": "Commits"},
            "config": {
                "padData": True,
                "showXAxis": True
            },
            "data": [
                { "key": "Commits", "color": "#9400D3", "values": values }
            ]
        }

        activity_js = self.j2_env.get_template('activity.js').render(
            commits_by_month=json.dumps(by_month),
            commits_by_year=json.dumps(by_year),
            recent_activity=json.dumps(recent_activity))
        with open(os.path.join(path, 'activity.js'), 'w') as fg:
            fg.write(activity_js)

        ###
        # Authors
        authors_html = self.render_authors_page()
        with open(os.path.join(path, "authors.html"), 'w', encoding='utf-8') as f:
            f.write(authors_html.decode('utf-8'))

        authors_activity_history = self.git_repository_statistics.authors.history(self._time_sampling_interval)

        authors_commits_history = self._squash_authors_history(authors_activity_history.commits_count,
                                                               self.configuration['max_authors'])

        authors_added_lines_history = self._squash_authors_history(authors_activity_history.insertions,
                                                                   self.configuration['max_authors'])

        # "Added lines" graph
        data = []

        for author in authors_added_lines_history:
            authorstats = {}
            authorstats['key'] = author
            series = authors_added_lines_history[author]
            authorstats['values'] = [{'x': int(x.timestamp()) * 1000, 'y': int(y)} for x, y in zip(series.index, series.values)]
            data.append(authorstats)

        lines_by_authors = {
            "xAxis": { "rotateLabels": -45 },
            "yAxis": { "axisLabel": "Lines" },
            "data" : data
        }

        # "Commit count" and streamgraph
        # TODO move the "added lines" into the same JSON to save space and download time
        data = []

        for author in authors_commits_history:
            authorstats = {}
            authorstats['key'] = author
            series = authors_commits_history[author]
            stream = series.diff().fillna(0)
            authorstats['values'] = [[int(x.timestamp() * 1000), int(y), int(z)] for x, y, z in zip(series.index, series.values, stream.values)]
            data.append(authorstats)

        commits_by_authors = {
            "xAxis": { "rotateLabels": -45 },
            "yAxis": { "axisLabel": "Commits" },
            "config": {
                "useInteractiveGuideline": True, 
                "style": "stream-center",
                "showControls": False,
                "showLegend": False,
            },
            "data": data
        }

        email_domains_distribution = self.git_repository_statistics.domains_distribution\
            .sort_values(ascending=False)
        if self.configuration['max_domains'] < email_domains_distribution.shape[0]:
            top_domains = email_domains_distribution[:self.configuration['max_domains']]
            other_domains = email_domains_distribution[self.configuration['max_domains']:].sum()
            email_domains_distribution = top_domains.append(pd.Series(other_domains, index=["Others"]))

        from collections import OrderedDict
        email_domains_distribution = email_domains_distribution.to_dict(OrderedDict)

        # Domains
        domains = {
            "config": {
                "donut": True,
                "padAngle": 0.01,
                "cornerRadius": 5
            },
            "data": [{"key": domain, "y": commits_count} for domain, commits_count in email_domains_distribution.items()]
        }

        authors_js = self.j2_env.get_template('authors.js').render(
            lines_by_authors=json.dumps(lines_by_authors),
            commits_by_authors=json.dumps(commits_by_authors),
            domains=json.dumps(domains)
        )
        with open(os.path.join(path, 'authors.js'), 'w') as fg:
            fg.write(authors_js)

        ###
        # Files
        files_html = self.render_files_page()

        with open(os.path.join(path, "files.html"), 'w', encoding='utf-8') as f:
            f.write(files_html)

        import pandas as pd
        hst = self.git_repository_statistics.linear_history(self._time_sampling_interval).copy()
        hst["epoch"] = (hst.index - pd.Timestamp("1970-01-01 00:00:00+00:00")) // pd.Timedelta('1s') * 1000

        files_count_ts = hst[["epoch", 'files_count']].rename(columns={"epoch": "x", 'files_count': "y"})\
            .to_dict('records')
        lines_count_ts = hst[["epoch", 'lines_count']].rename(columns={"epoch": "x", 'lines_count': "y"})\
            .to_dict('records')
        graph_data = {
            "xAxis": {"rotateLabels": -45},
            "yAxis1": {"axisLabel": "Files"},
            "yAxis2": {"axisLabel": "Lines"},
            "data": [
                {"key": "Files", "color": "#9400d3", "type": "line", "yAxis": 1, "values": files_count_ts},
                {"key": "Lines", "color": "#d30094", "type": "line", "yAxis": 2, "values": lines_count_ts},
            ]
        }

        files_js = self.j2_env.get_template('files.js').render(json_data=json.dumps(graph_data))
        with open(os.path.join(path, 'files.js'), 'w') as fg:
            fg.write(files_js)

        ###
        # tags.html

        if self.has_tags_page:
            tags_html = self.render_tags_page()
            with open(os.path.join(path, "tags.html"), 'w', encoding='utf-8') as f:
                f.write(tags_html.decode('utf-8'))

        ###
        # about.html
        about_html = self.render_about_page()
        with open(os.path.join(path, "about.html"), 'w', encoding='utf-8') as f:
            f.write(about_html.decode('utf-8'))

    def render_general_page(self):
        date_format_str = '%Y-%m-%d %H:%M'
        first_commit_datetime = datetime.datetime.fromtimestamp(self.git_repository_statistics.first_commit_timestamp)
        last_commit_datetime = datetime.datetime.fromtimestamp(self.git_repository_statistics.last_commit_timestamp)

        project_data = {
            "name": self.git_repo_statistics.repo_name,
            "branch": self.git_repo_statistics.analysed_branch,
            "age": (last_commit_datetime - first_commit_datetime).days,
            "active_days_count": self.git_repository_statistics.active_days_count,
            "commits_count": self.git_repository_statistics.total_commits_count,
            "authors_count": self.git_repository_statistics.authors.count(),
            "files_count": self.git_repo_statistics.total_files_count,
            "total_lines_count": self.git_repository_statistics.total_lines_count,
            "added_lines_count": self.git_repository_statistics.total_lines_added,
            "removed_lines_count": self.git_repository_statistics.total_lines_removed,
            "first_commit_date": first_commit_datetime.strftime(date_format_str),
            "last_commit_date": last_commit_datetime.strftime(date_format_str)
        }

        generation_data = {
            "datetime": datetime.datetime.today().strftime(date_format_str)
        }

        # load and render template
        template_rendered = self.j2_env.get_template('general.html').render(
            project=project_data,
            generation=generation_data,
            page_title="General",
            **self.common_rendering_data
        )
        return template_rendered

    def render_activity_page(self):
        # TODO: this conversion from old 'data' to new 'project data' should perhaps be removed in future
        project_data = {
            'timezones_activity': collections.OrderedDict(
                sorted(self.git_repository_statistics.timezones_distribution.items(), key=lambda n: int(n[0]))),
            'month_in_year_activity': self.git_repository_statistics.month_of_year_distribution.to_dict()
        }

        wd_h_distribution = self.git_repository_statistics.weekday_hour_distribution.astype('int32')
        project_data['weekday_hourly_activity'] = wd_h_distribution
        project_data['weekday_hour_max_commits_count'] = wd_h_distribution.max().max()
        project_data['weekday_activity'] = wd_h_distribution.sum(axis=1)
        project_data['hourly_activity'] = wd_h_distribution.sum(axis=0)

        # load and render template
        template_rendered = self.j2_env.get_template('activity.html').render(
            project=project_data,
            page_title="Activity",
            **self.common_rendering_data
        )
        return template_rendered

    def render_authors_page(self):
        project_data = {
            'top_authors': [],
            'non_top_authors': [],
            'authors_top': self.configuration['authors_top'],
            'total_commits_count': self.git_repository_statistics.total_commits_count,
            'total_lines_count': self.git_repository_statistics.total_lines_count
        }

        all_authors = self.git_repository_statistics.authors.sort().names()
        if len(all_authors) > self.configuration['max_authors']:
            project_data['non_top_authors'] = all_authors[self.configuration['max_authors']:]

        raw_authors_data = self.git_repository_statistics.get_authors_ranking_by_month()
        ordered_months = raw_authors_data.index.get_level_values(0).unique().sort_values(ascending=False)
        project_data['months'] = []
        for yymm in ordered_months[0:self.configuration['max_authors_of_months']]:
            authors_in_month = raw_authors_data.loc[yymm]
            project_data['months'].append({
                'date': yymm,
                'top_author': {'name': authors_in_month.index[0], 'commits_count': authors_in_month[0]},
                'next_top_authors': ', '.join(list(authors_in_month.index[1:5])),
                'all_commits_count': authors_in_month.sum(),
                'total_authors_count': authors_in_month.size
            })

        project_data['years'] = []
        raw_authors_data = self.git_repository_statistics.get_authors_ranking_by_year()
        max_top_authors_index = self.configuration['authors_top'] + 1
        ordered_years = raw_authors_data.index.get_level_values(0).unique().sort_values(ascending=False)
        for y in ordered_years:
            authors_in_year = raw_authors_data.loc[y]
            project_data['years'].append({
                'date': y,
                'top_author': {'name': authors_in_year.index[0], 'commits_count': authors_in_year[0]},
                'next_top_authors': ', '.join(list(authors_in_year.index[1:max_top_authors_index])),
                'all_commits_count': authors_in_year.sum(),
                'total_authors_count': authors_in_year.size
            })

        for author in all_authors[:self.configuration['max_authors']]:
            git_author = self.git_repository_statistics.get_author(author)
            author_dict = {
                'name': author,
                'commits_count': git_author.commits_count,
                'lines_added_count': git_author.lines_added,
                'lines_removed_count': git_author.lines_removed,
                'first_commit_date': git_author.first_commit_date.strftime('%Y-%m-%d'),
                'latest_commit_date': git_author.latest_commit_date.strftime('%Y-%m-%d'),
                'contributed_days_count': git_author.contributed_days_count,
                'active_days_count': git_author.active_days_count,
                'contribution': self.git_repo_statistics.contribution.get(author,
                                                                          0) if self.git_repo_statistics.contribution else None,
            }

            project_data['top_authors'].append(author_dict)

        # load and render template
        template_rendered = self.j2_env.get_template('authors.html').render(
            project=project_data,
            page_title="Authors",
            **self.common_rendering_data
        )
        return template_rendered.encode('utf-8')

    def render_files_page(self):
        # TODO: this conversion from old 'data' to new 'project data' should perhaps be removed in future
        project_data = {
            'files_count': self.git_repo_statistics.total_files_count,
            'lines_count': self.git_repository_statistics.total_lines_count,
            'size': self.git_repo_statistics.total_tree_size,
            'files': []
        }

        for ext in sorted(self.git_repo_statistics.extensions.keys()):
            files = self.git_repo_statistics.extensions[ext]['files']
            lines = self.git_repo_statistics.extensions[ext]['lines']
            file_type_dict = {"extension": ext,
                              "count": files,
                              "lines_count": lines
                              }
            project_data['files'].append(file_type_dict)

        # load and render template
        template_rendered = self.j2_env.get_template('files.html').render(
            project=project_data,
            page_title="Files",
            **self.common_rendering_data
        )
        return template_rendered

    def render_tags_page(self):
        # TODO: this conversion from old 'data' to new 'project data' should perhaps be removed in future
        project_data = {
            'tags_count': len(self.git_repo_statistics.tags),
            'tags': []
        }

        # TODO: fix error occurring when a tag name and project name are the same
        """
        fatal: ambiguous argument 'gitstats': both revision and filename
        Use '--' to separate paths from revisions, like this:
        'git <command> [<revision>...] -- [<file>...]'
        """
        tags_sorted_by_date_desc = sort_keys_by_value_of_key(self.git_repo_statistics.tags, 'date', reverse=True)
        for tag in tags_sorted_by_date_desc:
            if 'max_recent_tags' in self.configuration \
                    and self.configuration['max_recent_tags'] <= len(project_data['tags']):
                break
            # there are tags containing no commits
            if 'authors' in self.git_repo_statistics.tags[tag].keys():
                authordict = self.git_repo_statistics.tags[tag]['authors']
                authors_by_commits = [
                    name for name, _ in sorted(authordict.items(), key=lambda kv: kv[1], reverse=True)
                ]
                authorinfo = []
                for i in reversed(authors_by_commits):
                    authorinfo.append('%s (%d)' % (i, self.git_repo_statistics.tags[tag]['authors'][i]))
                tag_dict = {
                    'name': tag,
                    'date': self.git_repo_statistics.tags[tag]['date'],
                    'commits_count': self.git_repo_statistics.tags[tag]['commits'],
                    'authors': ', '.join(authorinfo)
                }
                project_data['tags'].append(tag_dict)

        # load and render template
        template_rendered = self.j2_env.get_template('tags.html').render(
            project=project_data,
            page_title="Tags",
            **self.common_rendering_data
        )
        return template_rendered.encode('utf-8')

    def render_about_page(self):
        repostat_version = self.configuration.get_release_data_info()['develop_version']
        repostat_version_date = self.configuration.get_release_data_info()['user_version']
        page_data = {
            "version": f"{repostat_version} ({repostat_version_date})",
            "tools": [GitStatistics.get_fetching_tool_info(),
                      self.configuration.get_jinja_version()],
            "contributors": [author for author in self.configuration.get_release_data_info()['contributors']]
        }

        template_rendered = self.j2_env.get_template('about.html').render(
            repostat=page_data,
            page_title="About",
            **self.common_rendering_data
        )
        return template_rendered.encode('utf-8')



import unittest

from analysis import GitStatistics
import test


class TestSingleCommitRepository(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.git_repository = test.GitRepository()

        cls.git_repository.commit_builder\
            .set_author("John Doe", "john@doe.com")\
            .add_file()\
            .commit()

        cls.commit_author, = cls.git_repository.commit_builder.author_signatures
        cls.gs = GitStatistics(cls.git_repository.location)

    def test_missing_authors_email(self):
        import tempfile
        import os
        from test import shellhelper

        cwd = os.getcwd()
        with tempfile.TemporaryDirectory() as tmpdir:
            # change working directory
            os.chdir(tmpdir)
            shellhelper.get_pipe_output(['git init'])
            shellhelper.get_pipe_output(['touch file.txt'])
            shellhelper.get_pipe_output(['git add file.txt'])
            # apparently it is not possible to create pygit.Signature with empty author's email (and name)
            # but commits with no author's email can be created via git
            shellhelper.get_pipe_output(['git commit -m "No-email author" --author "Author NoEmail <>"'])
            try:
                # Commit without author's email does not crash statistics calculation
                GitStatistics(tmpdir)
            except Exception as e:
                self.fail(str(e))
            os.chdir(cwd)



import unittest

from analysis import GitStatistics
import test


class TestTwoCommitRepository(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.git_repository = test.GitRepository()

        cls.git_repository.commit_builder\
            .set_author("John Doe I", "john1@doe.com")\
            .add_file(content=["Line by John Doe I"])\
            .commit()

        cls.git_repository.commit_builder\
            .set_author("John Doe II", "john2@doe.com")\
            .add_file(content=["Line 1 by John Doe II", "Line 2 by John Doe II"])\
            .commit()

    def test_contributors(self):
        gs = GitStatistics(self.git_repository.location)
        contributors = gs.fetch_contributors()
        # TODO: make commit builder remember the number of added lines and use info here
        self.assertDictEqual({'John Doe II': 2, 'John Doe I': 1}, contributors)



#!/usr/bin/env python3
# Copyright (c) 2007-2014 Heikki Hokkanen <hoxu@users.sf.net> & others
# Copyright (c) 2018-2020 Viktor Kopp <vifactor_at_gmail.com> & others
# GPLv2 / GPLv3
import os
import sys
import warnings
import time
import webbrowser

from report.htmlreportcreator import HTMLReportCreator
from analysis import GitStatistics
from analysis.gitrepository import GitRepository
from tools.configuration import Configuration

os.environ['LC_ALL'] = 'C'

time_start = time.time()


def get_execution_time():
    time_end = time.time()
    execution_time = time_end - time_start
    return execution_time


def main():
    try:
        config = Configuration(sys.argv[1:])
    except EnvironmentError as ee:
        warnings.warn("Environment exception occurred: {}".format(ee))
        sys.exit(1)

    print('Git path: %s' % config.git_repository_path)
    print('Collecting data...')
    repo_statistics = GitStatistics(config.git_repository_path,
                                    config.do_calculate_contribution(),
                                    config.do_process_tags())
    repository_statistics = GitRepository(config.git_repository_path)

    output_path = config.statistics_output_path
    print('Output path: %s' % output_path)
    os.makedirs(output_path, exist_ok=True)

    print('Generating HTML report...')
    HTMLReportCreator(config, repo_statistics, repository_statistics)\
        .set_time_sampling(config.get_time_sampling())\
        .create(output_path)
    exec_time_seconds = get_execution_time()
    print('Report generated in %.2f secs.' % exec_time_seconds)

    url = os.path.join(output_path, 'general.html').replace("'", "'\\''")
    if config.do_open_in_browser():
        webbrowser.open(url, new=2)
    else:
        print("You may open your report in a browser. Path: {}".format(url))


if __name__ == '__main__':
    main()



def sort_keys_by_value_of_key(d, key, reverse=False):
    return [el[1] for el in sorted([(d[el][key], el) for el in d.keys()], reverse=reverse)]


def split_email_address(email_address):
    parts = email_address.split('@')
    if len(parts) != 2:
        raise ValueError('Not an email passed: %s' % email_address)
    return parts[0], parts[1]



import os
import argparse
import json
import warnings

here = os.path.dirname(os.path.abspath(__file__))


class ReadableDir(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = os.path.abspath(os.path.expanduser(values))
        if not os.path.isdir(prospective_dir):
            raise argparse.ArgumentTypeError("Path {0} is not valid.".format(prospective_dir))
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            raise argparse.ArgumentTypeError("Directory {0} is not readable.".format(prospective_dir))


class WritableDir(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = os.path.abspath(os.path.expanduser(values))
        if os.path.isdir(prospective_dir):
            if not os.access(prospective_dir, os.W_OK):
                raise argparse.ArgumentTypeError("Directory {0} is not writable.".format(prospective_dir))
            warnings.warn("Directory {0} already exists. Its content will be rewritten.".format(prospective_dir))
        else:
            is_subdir_of_writable_parent = False
            parent_dir, sub_dir = os.path.split(prospective_dir)
            while sub_dir:
                if os.path.isdir(parent_dir) and os.access(parent_dir, os.W_OK):
                    is_subdir_of_writable_parent = True
                    break
                parent_dir, sub_dir = os.path.split(parent_dir)
            if not is_subdir_of_writable_parent:
                raise argparse.ArgumentTypeError("{0} is not writable directory.".format(parent_dir))

        setattr(namespace, self.dest, prospective_dir)


class ReadableFile(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        file_name = os.path.abspath(os.path.expanduser(values))
        if not os.path.exists(file_name):
            raise argparse.ArgumentTypeError("file:{0} does not exists".format(file_name))
        if not os.path.isfile(file_name):
            raise argparse.ArgumentTypeError("{0} is not a file".format(file_name))
        if not os.access(file_name, os.R_OK):
            raise argparse.ArgumentTypeError("file:{0} is not a readable".format(file_name))

        setattr(namespace, self.dest, file_name)


class Configuration(dict):
    release_data_dict = None

    @classmethod
    def get_release_data_info(cls):
        if not cls.release_data_dict:
            cls.release_data_dict = cls._read_release_data()
        return cls.release_data_dict

    @classmethod
    def _read_release_data(cls):
        release_data_file_path = os.path.join(here, 'release_data.json')
        with open(release_data_file_path) as release_json_file:
            return json.load(release_json_file)

    @classmethod
    def _read_config_data(cls, path):
        with open(path) as f:
            return json.load(f)

    def __init__(self, args_orig, **kwargs):
        dict.__init__(kwargs)

        self.args = self._parse_sys_argv(args_orig)
        self.git_repository_path = self.args.git_repo
        self.statistics_output_path = self.args.output_path

        self._set_default_configuration()
        if self.args.config_file:
            try:
                read_config = self._read_config_data(self.args.config_file)
                print("Read config", read_config)
                self.update(read_config)
            except Exception as e:
                print("Exception caught: ", e)

    def _set_default_configuration(self):
        self.update({
            "max_domains": 10,
            "max_ext_length": 10,
            "max_authors": 12,
            "max_authors_of_months": 6,
            "authors_top": 5,
            "colormap": "classic",
            "time_sampling": "W"
        })

    def do_open_in_browser(self):
        return not self.args.no_browser

    def is_report_relocatable(self):
        return self.args.copy_assets

    def do_calculate_contribution(self):
        return self.args.contribution

    def do_process_tags(self):
        return self["max_recent_tags"] > 0 if "max_recent_tags" in self else True

    def get_time_sampling(self):
        return self["time_sampling"] if "time_sampling" in self else "W"

    @classmethod
    def _parse_sys_argv(cls, argv):
        release_info = cls.get_release_data_info()
        parser = argparse.ArgumentParser(prog='repostat',
                                         description='Git repository desktop analyzer. '
                                                     'Analyze and generate git statistics '
                                                     'in HTML format')

        parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + release_info['develop_version'])
        parser.add_argument('-c', '--config-file', action=ReadableFile, help="Configuration file path")
        parser.add_argument('--no-browser', action="store_true", help="Do not open report in browser")
        parser.add_argument('--contribution', action="store_true", help="Add to html report contribution by author."
                                                                        "Attention: this is slow calculation!")
        parser.add_argument('--copy-assets', action="store_true",
                            help="Copy assets (images, css, etc.) into report folder (report becomes relocatable)")

        parser.add_argument('git_repo', type=str, action=ReadableDir, help="Path to git repository")
        parser.add_argument('output_path', type=str, action=WritableDir, help="Path to an output directory")

        return parser.parse_args(argv)

    @staticmethod
    def get_jinja_version():
        import jinja2 as j2
        return '{} v.{}'.format(j2.__name__, j2.__version__)



import time


class Timeit(object):

    def __init__(self, before_msg=None):
        self.before_message = before_msg

    def __call__(self, method):
        def wrapper(*args):
            if self.before_message is None:
                self.before_message = method.__name__
            print(self.before_message + "...")
            ts = time.time()
            result = method(*args)
            te = time.time()
            print('Finished in %2.2f ms' % ((te - ts) * 1000))
            return result
        return wrapper



import pygit2 as git
from datetime import datetime
import warnings
import os

from tools.timeit import Timeit


class GitStatistics:
    def __init__(self, path, fetch_contribution=False, fetch_tags=True):
        """
        :param path: path to a repository
        """
        self.repo = git.Repository(path)
        self.mailmap = git.Mailmap.from_repository(self.repo)

        self.repo_name = os.path.basename(os.path.abspath(path))
        self.analysed_branch = self.repo.head.shorthand

        if fetch_contribution:
            # this is slow
            self.contribution = self.fetch_contributors()
        else:
            self.contribution = {}
        if fetch_tags:
            self.tags = self.fetch_tags_info()
        else:
            self.tags = {}

        # extension -> files, lines, size
        self.extensions = self.get_current_files_info()
        self.total_files_count = sum(v['files'] for k, v in self.extensions.items())
        self.total_tree_size = sum(v['size'] for k, v in self.extensions.items())

    def map_signature(self, sig: git.Signature) -> git.Signature:
        """
        Maps of a contributor signature as read from a repository using a provided .mailmap file
        :param sig: Signature to map
        :return: mapped signature
        """
        try:
            mapped_signature = self.mailmap.resolve_signature(sig)
        except ValueError as e:
            name = sig.name
            email = sig.email
            if not name:
                name = "Empty Empty"
                warnings.warn(f"{str(e)}. Name will be replaced with '{name}'")
            if not email:
                email = "empty@empty.empty"
                warnings.warn(f"{str(e)}. Email will be replaced with '{email}'")
            return git.Signature(name, email, sig.time, sig.offset, 'utf-8')
        else:
            return mapped_signature

    @staticmethod
    def _get_file_extension(git_file_path, max_ext_length=5):
        filename = os.path.basename(git_file_path)
        basename_parts = filename.split('.')
        ext = basename_parts[1] if len(basename_parts) == 2 and basename_parts[0] else ''
        if len(ext) > max_ext_length:
            ext = ''
        return ext

    def get_current_files_info(self):
        """
        :return: returns total files count and distribution of lines and files count by file extensions
        """
        head_commit = self.repo.revparse_single('HEAD')
        head_commit_tree = head_commit.tree.diff_to_tree(swap=True)
        extensions = {}
        for p in head_commit_tree:
            ext = self._get_file_extension(p.delta.new_file.path)
            if ext not in extensions:
                extensions[ext] = {'files': 0, 'lines': 0, 'size': 0}
            _, lines_count, _ = p.line_stats
            extensions[ext]['lines'] += lines_count
            extensions[ext]['files'] += 1
            extensions[ext]['size'] += p.delta.new_file.size
        return extensions

    @classmethod
    def get_fetching_tool_info(cls):
        # could be bare git-subprocess invokation, PythonGit package, etc.
        return '{} v.{}'.format(git.__name__, git.LIBGIT2_VERSION)

    @Timeit("Fetching tags info")
    def fetch_tags_info(self):
        tags = [refobj for refobj in self.repo.listall_reference_objects() if refobj.name.startswith('refs/tags')]
        commit_tag = {refobj.peel().oid: refobj.shorthand for refobj in tags}

        result = {refobj.shorthand: {
            'stamp': refobj.peel().author.time,
            'date': datetime.fromtimestamp(refobj.peel().author.time).strftime('%Y-%m-%d'),
            'hash': str(refobj.target)} for refobj in tags}

        authors = {}
        commit_count = 0
        for commit in self.repo.walk(self.repo.head.target, git.GIT_SORT_TOPOLOGICAL | git.GIT_SORT_REVERSE):
            commit_count += 1
            commit_author = self.map_signature(commit.author)
            authors[commit_author.name] = authors.get(commit_author.name, 0) + 1
            if commit.oid in commit_tag.keys():
                tagname = commit_tag[commit.oid]
                result[tagname]['commits'] = commit_count
                result[tagname]['authors'] = authors

                commit_count = 0
                authors = {}

        return result

    @Timeit("Fetching current tree contributors")
    def fetch_contributors(self):
        head_commit = self.repo.head.peel()
        contribution = {}

        submodules_paths = self.repo.listall_submodules()
        diff_to_tree = head_commit.tree.diff_to_tree()
        diff_len = len(list(diff_to_tree))
        i = 0
        for p in diff_to_tree:
            file_to_blame = p.delta.new_file.path
            if file_to_blame not in submodules_paths and not p.delta.is_binary:
                blob_blame = self.repo.blame(file_to_blame)
                for blame_hunk in blob_blame:
                    hunk_committer = blame_hunk.final_committer
                    if not hunk_committer:
                        # if committer configured an empty email when created commit
                        # blame hunk corresponding to that commit will produce a None signature
                        # the following substitutes hunk's final committer with an author of the commit
                        hunk_committer = self.repo[blame_hunk.orig_commit_id].author
                    committer = self.map_signature(hunk_committer)
                    contribution[committer.name] = contribution.get(committer.name, 0) + blame_hunk.lines_in_hunk
            i += 1
            print(f"Working... ({i} / {diff_len})", end="\r", flush=True)

        return contribution



import unittest
import os
from pygit2 import Signature

import test
from analysis.gitdata import WholeHistory, LinearHistory


class GitHistoryTest(unittest.TestCase):
    def setUp(self):
        self.test_repo = test.GitRepository()

        # create commit in master branch (created by default)
        self.test_repo.commit_builder \
            .set_author("John Doe", "john@doe.com") \
            .add_file(content="bzyk") \
            .commit()

    def test_records_count_in_history(self):
        # create second branch from 'master' (which already had one commit)
        branch = self.test_repo.branches.local.create('second_branch', self.test_repo.head.peel())
        self.test_repo.checkout(branch)

        # create commit on new branch
        self.test_repo.commit_builder \
            .set_author("Author Author", "author@author.net") \
            .add_file(content=["some content"]) \
            .commit()

        # so far no merge commits, both linear and whole history caches should contain 2 records
        whole_history_df = WholeHistory(self.test_repo).as_dataframe()
        self.assertEqual(2, len(whole_history_df.index))
        linear_history_df = LinearHistory(self.test_repo).as_dataframe()
        self.assertEqual(2, len(linear_history_df.index))

        # now merge commit is being created
        # checkout to master
        master_branch = self.test_repo.branches.get('master')
        self.test_repo.checkout(master_branch)
        # and merge 'second_branch' into 'master'
        self.test_repo.merge(self.test_repo.branches.get('second_branch').peel().id)
        # by creating merge commit
        author = Signature("name", "email")
        committer = author
        tree = self.test_repo.index.write_tree()
        message = "Merge 'second_branch' into 'master'"
        self.test_repo.create_commit('HEAD', author, committer, message, tree,
                                     [self.test_repo.head.target,
                                      self.test_repo.branches.get('second_branch').peel().oid])

        # whole history cache should contain 3 records: initial commit + commit in merged branch + merge commit
        whole_history_df = WholeHistory(self.test_repo).as_dataframe()
        self.assertEqual(3, len(whole_history_df.index))
        # linear history cache still contains 2 records: initial commit + merge commit
        linear_history_df = LinearHistory(self.test_repo).as_dataframe()
        self.assertEqual(2, len(linear_history_df.index))

    def test_mailmap(self):
        # create second commit with another signature
        self.test_repo.commit_builder \
            .set_author("Author Author", "author@author.net") \
            .add_file(content=["some content"]) \
            .commit()

        # map second author's signature as first one using mailmap file
        commit_author1, commit_author2 = self.test_repo.commit_builder.author_signatures
        with open(os.path.join(self.test_repo.location, ".mailmap"), 'w') as mm:
            mm.write(f"{commit_author1.name} <{commit_author1.email}> "
                     f"{commit_author2.name} <{commit_author2.email}>")

        # Mailmap might be tested either on Whole or Linear history
        wh_df = WholeHistory(self.test_repo).as_dataframe()

        authors = wh_df['author_name'].unique()
        self.assertCountEqual(["John Doe"], authors)
        emails = wh_df['author_email'].values
        # both emails are preserved for statistics
        self.assertCountEqual(["john@doe.com", "author@author.net"], emails)

    # TODO: add test for inserted/deleted lines count



import pandas as pd


class GitAuthors(object):
    def __init__(self, git_history: pd.DataFrame):
        self.raw_authors_data = git_history[['author_timestamp', 'author_name', 'insertions', 'deletions']]

        authors_grouped = self.raw_authors_data[['author_name', 'insertions', 'deletions']].groupby(
            [self.raw_authors_data['author_name']])
        self.authors_summary = authors_grouped.sum()
        self.authors_summary['commits_count'] = authors_grouped['author_name'].count()
        self.authors_summary.reset_index(inplace=True)

    def count(self):
        return self.authors_summary.shape[0]

    def names(self):
        return self.authors_summary['author_name'].values

    def sort(self, by="commits_count"):
        self.authors_summary = self.authors_summary.sort_values(by=by, ascending=False)
        return self

    def get(self, name):
        # TODO: this should produce GitAuthor object
        return self.authors_summary.loc[self.authors_summary['author_name'] == name]

    @property
    def summary(self):
        return self.authors_summary

    def history(self, sampling):
        """
        :param sampling: frequency string https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        :return: Dataframe with Datetime multiindexindex and changes history data by authors
        """
        wh = self.raw_authors_data[['author_name', 'insertions', 'deletions']].copy()
        wh['author_datetime'] = pd.to_datetime(self.raw_authors_data['author_timestamp'], unit='s', utc=True)
        wh = wh.set_index(wh['author_datetime'])
        wh_grouped = wh[['author_name', 'insertions', 'deletions']].groupby(
            [wh['author_name'], pd.Grouper(freq=sampling)])

        modifications_over_time = wh_grouped.sum()\
            .reset_index()

        commits_over_time = wh_grouped.count().rename(columns={'author_name': 'commits_count'})\
            .reset_index()

        modifications_over_time['commits_count'] = commits_over_time['commits_count']

        modifications_per_author_over_time = modifications_over_time.reset_index().pivot_table(
            index=modifications_over_time['author_datetime'],
            columns=modifications_over_time['author_name'],
            values=['insertions', 'deletions', 'commits_count']).fillna(0)
        return modifications_per_author_over_time
